/*

  Copyright (C) 2000,2002,2004,2005,2006 Silicon Graphics, Inc.  All Rights Reserved.

  Copyright (C) 2006  NEC Corporation
  This file was modified from dwarfdump to analyze kernel debug information
  in makedumpfile.
  See below for the original copyright notice.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2.1 of the GNU Lesser General Public License 
  as published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement 
  or the like.  Any license provided herein, whether implied or 
  otherwise, applies only to this software file.  Patent licenses, if
  any, provided herein do not apply to combinations of this program with 
  other software, or any other product whatsoever.  

  You should have received a copy of the GNU Lesser General Public 
  License along with this program; if not, write the Free Software 
  Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, 
  USA.

  Contact information:  Silicon Graphics, Inc., 1500 Crittenden Lane,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan

*/



#include "config.h"
#include "dwarf_incl.h"
#include <stdio.h>
#include <stdlib.h>
#include "dwarf_line.h"
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif


/* 
    Although source files is supposed to return the
    source files in the compilation-unit, it does
    not look for any in the statement program.  In
    other words, it ignores those defined using the
    extended opcode DW_LNE_define_file.
*/
int
dwarf_srcfiles(Dwarf_Die die,
	       char ***srcfiles,
	       Dwarf_Signed * srcfilecount, Dwarf_Error * error)
{
    /* 
       This pointer is used to scan the portion of the .debug_line
       section for the current cu. */
    Dwarf_Small *line_ptr;


    /* 
       Pointer to a DW_AT_stmt_list attribute in case it exists in the
       die. */
    Dwarf_Attribute stmt_list_attr;

    /* Pointer to name of compilation directory. */
    Dwarf_Small *comp_dir = 0;

    /* Offset into .debug_line specified by a DW_AT_stmt_list
       attribute. */
    Dwarf_Unsigned line_offset = 0;



    /* 
       This points to a block of char *'s, each of which points to a
       file name. */
    char **ret_files = 0;

    /* The Dwarf_Debug this die belongs to. */
    Dwarf_Debug dbg = 0;

    /* Used to chain the file names. */
    Dwarf_Chain curr_chain, prev_chain = NULL, head_chain = NULL;
    int resattr = 0;
    int lres = 0;
    struct Line_Table_Prefix_s line_prefix;
    int i = 0;
    int res = 0;


    /* ***** BEGIN CODE ***** */

    /* Reset error. */
    if (error != NULL)
	*error = NULL;

    CHECK_DIE(die, DW_DLV_ERROR)
	dbg = die->di_cu_context->cc_dbg;

    resattr = dwarf_attr(die, DW_AT_stmt_list, &stmt_list_attr, error);
    if (resattr != DW_DLV_OK) {
	return resattr;
    }

    if (dbg->de_debug_line_index == 0) {
	_dwarf_error(dbg, error, DW_DLE_DEBUG_LINE_NULL);
	return (DW_DLV_ERROR);
    }

    res =
	_dwarf_load_section(dbg,
			    dbg->de_debug_line_index,
			    &dbg->de_debug_line, error);
    if (res != DW_DLV_OK) {
	return res;
    }

    lres = dwarf_formudata(stmt_list_attr, &line_offset, error);
    if (lres != DW_DLV_OK) {
	return lres;
    }
    if (line_offset >= dbg->de_debug_line_size) {
	_dwarf_error(dbg, error, DW_DLE_LINE_OFFSET_BAD);
	return (DW_DLV_ERROR);
    }
    line_ptr = dbg->de_debug_line + line_offset;
    dwarf_dealloc(dbg, stmt_list_attr, DW_DLA_ATTR);

    dwarf_init_line_table_prefix(&line_prefix);
    {
	Dwarf_Small *line_ptr_out = 0;
	int dres = dwarf_read_line_table_prefix(dbg,
						line_ptr,
						dbg->de_debug_line_size,
						&line_ptr_out,
						&line_prefix, error);

	if (dres == DW_DLV_ERROR)
	    return dres;
	if (dres == DW_DLV_NO_ENTRY) {
	    dwarf_free_line_table_prefix(&line_prefix);
	    return dres;
	}

	line_ptr = line_ptr_out;
    }

    for (i = 0; i < line_prefix.pf_files_count; ++i) {
	struct Line_Table_File_Entry_s *fe =
	    line_prefix.pf_line_table_file_entries + i;
	char *file_name = (char *) fe->lte_filename;
	char *dir_name = 0;
	char *full_name = 0;
	Dwarf_Unsigned dir_index = fe->lte_directory_index;

	if (dir_index == 0) {
	    dir_name = (char *) comp_dir;
	} else {
	    dir_name =
		(char *) line_prefix.pf_include_directories[fe->
							    lte_directory_index
							    - 1];
	}

	/* dir_name can be NULL if there is no DW_AT_comp_dir */
	if ((*file_name) == '/' || dir_name == 0) {
	    /* This is safe because dwarf_dealloc is careful to not
	       dealloc strings which are part of the raw .debug_* data. 
	     */
	    full_name = file_name;
	} else {
	    full_name = (char *) _dwarf_get_alloc(dbg, DW_DLA_STRING,
						  strlen(dir_name) + 1 +
						  strlen(file_name) +
						  1);
	    if (full_name == NULL) {
		_dwarf_error(dbg, error, DW_DLE_ALLOC_FAIL);
		return (DW_DLV_ERROR);
	    }

	    /* This is not careful to avoid // in the output, Nothing
	       forces a 'canonical' name format here. Unclear if this
	       needs to be fixed. */
	    strcpy(full_name, dir_name);
	    strcat(full_name, "/");
	    strcat(full_name, file_name);
	}
	curr_chain =
	    (Dwarf_Chain) _dwarf_get_alloc(dbg, DW_DLA_CHAIN, 1);
	if (curr_chain == NULL) {
	    _dwarf_error(dbg, error, DW_DLE_ALLOC_FAIL);
	    return (DW_DLV_ERROR);
	}
	curr_chain->ch_item = full_name;
	if (head_chain == NULL)
	    head_chain = prev_chain = curr_chain;
	else {
	    prev_chain->ch_next = curr_chain;
	    prev_chain = curr_chain;
	}


    }

    curr_chain = (Dwarf_Chain) _dwarf_get_alloc(dbg, DW_DLA_CHAIN, 1);
    if (curr_chain == NULL) {
	_dwarf_error(dbg, error, DW_DLE_ALLOC_FAIL);
	return (DW_DLV_ERROR);
    }




    if (line_prefix.pf_files_count == 0) {
	*srcfiles = NULL;
	*srcfilecount = 0;
	return (DW_DLV_NO_ENTRY);
    }

    ret_files = (char **)
	_dwarf_get_alloc(dbg, DW_DLA_LIST, line_prefix.pf_files_count);
    if (ret_files == NULL) {
	_dwarf_error(dbg, error, DW_DLE_ALLOC_FAIL);
	return (DW_DLV_ERROR);
    }

    curr_chain = head_chain;
    for (i = 0; i < line_prefix.pf_files_count; i++) {
	*(ret_files + i) = curr_chain->ch_item;
	prev_chain = curr_chain;
	curr_chain = curr_chain->ch_next;
	dwarf_dealloc(dbg, prev_chain, DW_DLA_CHAIN);
    }

    *srcfiles = ret_files;
    *srcfilecount = line_prefix.pf_files_count;
    return (DW_DLV_OK);
}

/* Operand counts per standard operand.
   The initial zero is for DW_LNS_copy. 
   This is an economical way to verify we understand the table
   of standard-opcode-lengths in the line table prologue.  */
#define STANDARD_OPERAND_COUNT_DWARF2 9
#define STANDARD_OPERAND_COUNT_DWARF3 12
static unsigned char
  dwarf_standard_opcode_operand_count[STANDARD_OPERAND_COUNT_DWARF3] = {
    /* DWARF2 */
    0,
    1, 1, 1, 1,
    0, 0, 0,
    1,
    /* Following are new for DWARF3. */
    0, 0, 1
};

/* Common line table prefix reading code. 
   Returns DW_DLV_OK, DW_DLV_ERROR.
   DW_DLV_NO_ENTRY cannot be returned, but callers should
   assume it is possible.

   The prefix_out area must be initialized properly before calling this.

   Has the side effect of allocating arrays which
   must be freed (see the Line_Table_Prefix_s struct which
   holds the pointers to space we allocate here).
*/
int
dwarf_read_line_table_prefix(Dwarf_Debug dbg,
			     Dwarf_Small * data_start,
			     Dwarf_Unsigned data_length,
			     Dwarf_Small ** updated_data_start_out,
			     struct Line_Table_Prefix_s *prefix_out,
			     Dwarf_Error * err)
{
    Dwarf_Small *line_ptr = data_start;
    Dwarf_Unsigned total_length = 0;
    int local_length_size = 0;
    int local_extension_size = 0;
    Dwarf_Unsigned prologue_length = 0;
    Dwarf_Half version = 0;
    Dwarf_Unsigned directories_count = 0;
    Dwarf_Unsigned directories_malloc = 0;
    Dwarf_Unsigned files_count = 0;
    Dwarf_Unsigned files_malloc = 0;
    Dwarf_Small *line_ptr_end = 0;

    prefix_out->pf_line_ptr_start = line_ptr;
    /* READ_AREA_LENGTH updates line_ptr for consumed bytes */
    READ_AREA_LENGTH(dbg, total_length, Dwarf_Unsigned,
		     line_ptr, local_length_size, local_extension_size);


    line_ptr_end = line_ptr + total_length;
    prefix_out->pf_line_ptr_end = line_ptr_end;
    prefix_out->pf_length_field_length = local_length_size +
	local_extension_size;
    /* ASSERT: prefix_out->pf_length_field_length == line_ptr
       -prefix_out->pf_line_ptr_start; */
    if (line_ptr_end > dbg->de_debug_line + dbg->de_debug_line_size) {
	_dwarf_error(dbg, err, DW_DLE_DEBUG_LINE_LENGTH_BAD);
	return (DW_DLV_ERROR);
    }
    if (line_ptr_end > data_start + data_length) {
	_dwarf_error(dbg, err, DW_DLE_DEBUG_LINE_LENGTH_BAD);
	return (DW_DLV_ERROR);
    }
    prefix_out->pf_total_length = total_length;

    READ_UNALIGNED(dbg, version, Dwarf_Half,
		   line_ptr, sizeof(Dwarf_Half));
    prefix_out->pf_version = version;
    line_ptr += sizeof(Dwarf_Half);
    if (version != CURRENT_VERSION_STAMP &&
	version != CURRENT_VERSION_STAMP3) {
	_dwarf_error(dbg, err, DW_DLE_VERSION_STAMP_ERROR);
	return (DW_DLV_ERROR);
    }

    READ_UNALIGNED(dbg, prologue_length, Dwarf_Unsigned,
		   line_ptr, local_length_size);
    prefix_out->pf_prologue_length = prologue_length;
    line_ptr += local_length_size;
    prefix_out->pf_line_prologue_start = line_ptr;


    prefix_out->pf_minimum_instruction_length =
	*(unsigned char *) line_ptr;
    line_ptr = line_ptr + sizeof(Dwarf_Small);

    prefix_out->pf_default_is_stmt = *(unsigned char *) line_ptr;
    line_ptr = line_ptr + sizeof(Dwarf_Small);

    prefix_out->pf_line_base = *(signed char *) line_ptr;
    line_ptr = line_ptr + sizeof(Dwarf_Sbyte);

    prefix_out->pf_line_range = *(unsigned char *) line_ptr;
    line_ptr = line_ptr + sizeof(Dwarf_Small);

    prefix_out->pf_opcode_base = *(unsigned char *) line_ptr;
    line_ptr = line_ptr + sizeof(Dwarf_Small);

    /* Set up the array of standard opcode lengths. */
    /* We think this works ok even for cross-endian processing of
       objects.  It might be wrong, we might need to specially process
       the array of ubyte into host order.  */
    prefix_out->pf_opcode_length_table = line_ptr;

    /* pf_opcode_base is one greater than the size of the array. */
    line_ptr += prefix_out->pf_opcode_base - 1;

    {
	/* Determine (as best we can) whether the
	   pf_opcode_length_table holds 9 or 12 standard-conforming
	   entries.  gcc4 upped to DWARF3's 12 without updating the
	   version number.  */
	int operand_ck_fail = true;

	if (prefix_out->pf_opcode_base >= STANDARD_OPERAND_COUNT_DWARF3) {
	    int mismatch = memcmp(dwarf_standard_opcode_operand_count,
				  prefix_out->pf_opcode_length_table,
				  STANDARD_OPERAND_COUNT_DWARF3);

	    if (!mismatch) {
		operand_ck_fail = false;
		prefix_out->pf_std_op_count =
		    STANDARD_OPERAND_COUNT_DWARF3;
	    }

	}
	if (operand_ck_fail) {
	    if (prefix_out->pf_opcode_base >=
		STANDARD_OPERAND_COUNT_DWARF2) {

		int mismatch =
		    memcmp(dwarf_standard_opcode_operand_count,
			   prefix_out->pf_opcode_length_table,
			   STANDARD_OPERAND_COUNT_DWARF2);

		if (!mismatch) {
		    operand_ck_fail = false;
		    prefix_out->pf_std_op_count =
			STANDARD_OPERAND_COUNT_DWARF2;
		}
	    }
	}
	if (operand_ck_fail) {
	    _dwarf_error(dbg, err, DW_DLE_LINE_NUM_OPERANDS_BAD);
	    return (DW_DLV_ERROR);
	}
    }
    /* At this point we no longer need to check operand counts. */


    directories_count = 0;
    directories_malloc = 5;
    prefix_out->pf_include_directories = malloc(sizeof(Dwarf_Small *) *
						directories_malloc);
    if (prefix_out->pf_include_directories == NULL) {
	_dwarf_error(dbg, err, DW_DLE_ALLOC_FAIL);
	return (DW_DLV_ERROR);
    }
    memset(prefix_out->pf_include_directories, 0,
	   sizeof(Dwarf_Small *) * directories_malloc);

    while ((*(char *) line_ptr) != '\0') {
	if (directories_count >= directories_malloc) {
	    Dwarf_Unsigned expand = 2 * directories_malloc;
	    Dwarf_Unsigned bytesalloc = sizeof(Dwarf_Small *) * expand;
	    Dwarf_Small **newdirs =
		realloc(prefix_out->pf_include_directories,
			bytesalloc);

	    if (!newdirs) {
		_dwarf_error(dbg, err, DW_DLE_ALLOC_FAIL);
		return (DW_DLV_ERROR);
	    }
	    /* Doubled size, zero out second half. */
	    memset(newdirs + directories_malloc, 0,
		   sizeof(Dwarf_Small *) * directories_malloc);
	    directories_malloc = expand;
	    prefix_out->pf_include_directories = newdirs;
	}
	prefix_out->pf_include_directories[directories_count] =
	    line_ptr;
	line_ptr = line_ptr + strlen((char *) line_ptr) + 1;
	directories_count++;
    }
    prefix_out->pf_include_directories_count = directories_count;
    line_ptr++;

    files_count = 0;
    files_malloc = 5;
    prefix_out->pf_line_table_file_entries =
	malloc(sizeof(struct Line_Table_File_Entry_s) * files_malloc);
    if (prefix_out->pf_line_table_file_entries == NULL) {
	_dwarf_error(dbg, err, DW_DLE_ALLOC_FAIL);
	return (DW_DLV_ERROR);
    }
    memset(prefix_out->pf_line_table_file_entries, 0,
	   sizeof(struct Line_Table_File_Entry_s) * files_malloc);

    while (*(char *) line_ptr != '\0') {
	Dwarf_Unsigned utmp;
	Dwarf_Unsigned dir_index = 0;
	Dwarf_Unsigned lastmod = 0;
	Dwarf_Unsigned file_length = 0;
	struct Line_Table_File_Entry_s *curline;
	Dwarf_Word leb128_length = 0;


	if (files_count >= files_malloc) {
	    Dwarf_Unsigned expand = 2 * files_malloc;
	    struct Line_Table_File_Entry_s *newfiles =
		realloc(prefix_out->pf_line_table_file_entries,
			sizeof(struct Line_Table_File_Entry_s) *
			expand);
	    if (!newfiles) {
		_dwarf_error(dbg, err, DW_DLE_ALLOC_FAIL);
		return (DW_DLV_ERROR);
	    }
	    memset(newfiles + files_malloc, 0,
		   sizeof(struct Line_Table_File_Entry_s) *
		   files_malloc);
	    files_malloc = expand;
	    prefix_out->pf_line_table_file_entries = newfiles;
	}
	curline = prefix_out->pf_line_table_file_entries + files_count;

	curline->lte_filename = line_ptr;
	line_ptr = line_ptr + strlen((char *) line_ptr) + 1;

	DECODE_LEB128_UWORD(line_ptr, utmp)
	    dir_index = (Dwarf_Sword) utmp;
	if (dir_index > directories_count) {
	    _dwarf_error(dbg, err, DW_DLE_DIR_INDEX_BAD);
	    return (DW_DLV_ERROR);
	}
	curline->lte_directory_index = dir_index;

	lastmod = _dwarf_decode_u_leb128(line_ptr, &leb128_length);
	line_ptr = line_ptr + leb128_length;
	curline->lte_last_modification_time = lastmod;

	/* Skip over file length. */
	file_length = _dwarf_decode_u_leb128(line_ptr, &leb128_length);
	line_ptr = line_ptr + leb128_length;
	curline->lte_length_of_file = file_length;

	++files_count;

    }
    prefix_out->pf_files_count = files_count;
    /* Skip trailing nul byte */
    ++line_ptr;


    if (line_ptr != (prefix_out->pf_line_prologue_start +
		     prefix_out->pf_prologue_length)) {
	_dwarf_error(dbg, err, DW_DLE_LINE_PROLOG_LENGTH_BAD);
	return (DW_DLV_ERROR);
    }


    *updated_data_start_out = line_ptr;
    return DW_DLV_OK;
}

/* Initialize the Line_Table_Prefix_s struct. 
   memset is not guaranteed a portable initializer, but works
   fine for current architectures.   AFAIK.
*/
void
dwarf_init_line_table_prefix(struct Line_Table_Prefix_s *pf)
{
    memset(pf, 0, sizeof(*pf));
}

/* Free any malloc'd area.  of the Line_Table_Prefix_s struct. */
void
dwarf_free_line_table_prefix(struct Line_Table_Prefix_s *pf)
{
    if (pf->pf_include_directories) {
	free(pf->pf_include_directories);
	pf->pf_include_directories = 0;
    }
    if (pf->pf_line_table_file_entries) {
	free(pf->pf_line_table_file_entries);
	pf->pf_line_table_file_entries = 0;
    }
    return;
}
