/* 
  Copyright (C) 2000,2002,2004,2005 Silicon Graphics, Inc.  All Rights Reserved.

  Copyright (C) 2006  NEC Corporation
  This file was modified from dwarfdump to analyze kernel debug information
  in makedumpfile.
  See below for the original copyright notice.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement
  or the like.  Any license provided herein, whether implied or
  otherwise, applies only to this software file.  Patent licenses, if
  any, provided herein do not apply to combinations of this program with
  other software, or any other product whatsoever.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston MA 02111-1307, USA.

  Contact information:  Silicon Graphics, Inc., 1500 Crittenden Lane,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan


$Header: /plroot/cmplrs.src/v7.4.5m/.RCS/PL/dwarfdump/RCS/dwarfdump.c,v 1.48 2006/04/18 18:05:57 davea Exp $ */
#include "globals.h"

/* for 'open' */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "dwconf.h"
#include "dwarf_incl.h"
extern char *optarg;

#define COMPLETED 0
#define FAILED 1
#define BYTES_PER_INSTRUCTION 4

static void print_infos(Dwarf_Debug dbg);

static string program_name = CMDNAME;
int check_error = 0;

/* defined in print_sections.c, die for the current compile unit, 
   used in get_fde_proc_name() */


#ifdef DEBUG
boolean ellipsis = FALSE;
#endif

Dwarf_Error err;
jmp_buf jmpbuf;

static int process_one_file(Elf * elf, int archive);

struct dwarf_info dwarf_info;

int
open_kernel_file(string name)
{
    int f = 0;

#ifdef __CYGWIN__
    f = open(name, O_RDONLY | O_BINARY);
#else
    f = open(name, O_RDONLY);
#endif
    return f;

}

#ifdef DEBUG
/*
 * Iterate through dwarf and print all info.
 */
char **argv_main;

int
main(int argc, char *argv[])
{

/* for debug */
	argv_main = argv;
	dwarf_info.status = 0;
	dwarf_info.struct_name = "page";
	dwarf_info.file_name = "mm.h";
	dwarf_info.struct_size = 0;
	dwarf_info.no_of_member = 3;
	dwarf_info.cur_member = DWARF_INFO_MAX;
	dwarf_info.member[0].name = "flags";
	dwarf_info.member[0].offset = 0;
	dwarf_info.member[1].name = "_count";
	dwarf_info.member[1].offset = 0;

	if (argc >= 3 && !strcmp(argv[2], "15")) {
		dwarf_info.location_member = DWARF_INFO_MAX;
		dwarf_info.member[2].name = "mapping";
		dwarf_info.member[2].offset = 0;
	} else {
		dwarf_info.location_member = 2;
		dwarf_info.member[2].name = NULL; /* has no name */
		dwarf_info.member[2].offset = 0;
	}

	dwarfinfo();
}
#endif

int
dwarfinfo()
{
    Elf_Cmd cmd;
    Elf *arf, *elf;
    int archive = 0;

    if (setjmp(jmpbuf)) {
	if (check_error)
	    return FAILED;
	else
	    return COMPLETED;
    }

    (void) elf_version(EV_NONE);
    if (elf_version(EV_CURRENT) == EV_NONE) {
	(void) fprintf(stderr, "%s: libelf.a out of date.\n", CMDNAME);
	return FAILED;
    }

#ifdef DEBUG
    string file_name = argv_main[1];
    dwarf_info.vmlinux_name = file_name;
    dwarf_info.vmlinux_fd = open_kernel_file(file_name);
    if (dwarf_info.vmlinux_fd == -1) {
	fprintf(stderr, "Can't open %s\n",
		file_name);
	return (FAILED);
    }
#endif

    cmd = ELF_C_READ;
    arf = elf_begin(dwarf_info.vmlinux_fd, cmd, (Elf *) 0);
    if (elf_kind(arf) == ELF_K_AR) {
	archive = 1;
    }
    while ((elf = elf_begin(dwarf_info.vmlinux_fd, cmd, arf)) != 0) {
	Elf32_Ehdr *eh32;

#ifdef HAVE_ELF64_GETEHDR
	Elf64_Ehdr *eh64;
#endif /* HAVE_ELF64_GETEHDR */
	eh32 = elf32_getehdr(elf);
	if (!eh32) {
#ifdef HAVE_ELF64_GETEHDR
	    /* not a 32-bit obj */
	    eh64 = elf64_getehdr(elf);
	    if (!eh64) {
		/* not a 64-bit obj either! */
		/* Quiet when not an object. */
	    } else {
		process_one_file(elf, archive);
	    }
#endif /* HAVE_ELF64_GETEHDR */
	} else {
	    process_one_file(elf, archive);
	}
	if (check_error)
	    break;
	cmd = elf_next(elf);
	elf_end(elf);
	if (dwarf_info.status & DWARF_INFO_DONE)
	    break;
    }
    elf_end(arf);
#ifdef DEBUG
/* for debug */
	write (1, &dwarf_info, sizeof(struct dwarf_info));
#endif
    if (check_error)
	return FAILED;
    else
	return COMPLETED;
}

/*
  Given a file which we know is an elf file, process
  the dwarf data.

*/
static int
process_one_file(Elf * elf, int archive)
{
    Dwarf_Debug dbg;
    int dres;

    dres = dwarf_elf_init(elf, DW_DLC_READ, NULL, NULL, &dbg, &err);
    if (dres == DW_DLV_NO_ENTRY) {
	fprintf(stderr, "No DWARF information present in %s\n", dwarf_info.vmlinux_name);
	check_error = 1;
	return 0;
    }
    if (dres != DW_DLV_OK) {
	print_error(dbg, "dwarf_elf_init", dres, err);
    }

#ifdef DEBUG
    if (archive) {
	Elf_Arhdr *mem_header = elf_getarhdr(elf);

	fprintf(stderr, "\narchive member \t%s\n",
	       mem_header ? mem_header->ar_name : "");
    }
#endif
	print_infos(dbg);
	dres = dwarf_finish(dbg, &err);
    if (dres != DW_DLV_OK) {
	print_error(dbg, "dwarf_finish", dres, err);
    }
    return 0;

}

/* process each compilation unit in .debug_info */
static void
print_infos(Dwarf_Debug dbg)
{
    Dwarf_Unsigned cu_header_length = 0;
    Dwarf_Unsigned abbrev_offset = 0;
    Dwarf_Half version_stamp = 0;
    Dwarf_Half address_size = 0;
    Dwarf_Die cu_die = 0;
    Dwarf_Unsigned next_cu_offset = 0;
    int nres = DW_DLV_OK;

	/* The followings are for debug only. SAKUJO */
#ifdef DEBUG
    fprintf(stderr, "\n.debug_info\n");
#endif

    /* Loop until it fails.  */
    while ((nres =
	    dwarf_next_cu_header(dbg, &cu_header_length, &version_stamp,
				 &abbrev_offset, &address_size,
				 &next_cu_offset, &err))
	   == DW_DLV_OK) {
	int sres;


	/* process a single compilation unit in .debug_info. */
	sres = dwarf_siblingof(dbg, NULL, &cu_die, &err);
	if (sres == DW_DLV_OK) {
		Dwarf_Signed cnt = 0;
		char **srcfiles = 0;
		int srcf = dwarf_srcfiles(cu_die,
					  &srcfiles, &cnt, &err);

		if (srcf != DW_DLV_OK) {
		    srcfiles = 0;
		    cnt = 0;
		}

		print_die_and_children(dbg, cu_die, srcfiles, cnt);
		if (srcf == DW_DLV_OK) {
		    int si;

		    for (si = 0; si < cnt; ++si) {
			dwarf_dealloc(dbg, srcfiles[si], DW_DLA_STRING);
		    }
		    dwarf_dealloc(dbg, srcfiles, DW_DLA_LIST);
		}
	    dwarf_dealloc(dbg, cu_die, DW_DLA_DIE);
	} else if (sres == DW_DLV_NO_ENTRY) {
	    /* do nothing I guess. */
	} else {
	    print_error(dbg, "Regetting cu_die", sres, err);
	}
	if (dwarf_info.status & DWARF_INFO_DONE)
	    break;
    }
    if (nres == DW_DLV_ERROR) {
	check_error = 1;
	string errmsg = dwarf_errmsg(err);
	Dwarf_Unsigned myerr = dwarf_errno(err);

	fprintf(stderr, "%s ERROR:  %s:  %s (%lu)\n",
		program_name, "attempting to print .debug_info",
		errmsg, (unsigned long) myerr);
	fprintf(stderr, "attempting to continue.\n");
    }
}

/* ARGSUSED */
void
print_error(Dwarf_Debug dbg, string msg, int dwarf_code,
	    Dwarf_Error err)
{
    if (dwarf_code == DW_DLV_ERROR) {
	string errmsg = dwarf_errmsg(err);
	Dwarf_Unsigned myerr = dwarf_errno(err);

	fprintf(stderr, "%s ERROR:  %s:  %s (%lu)\n",
		program_name, msg, errmsg, (unsigned long) myerr);
    } else if (dwarf_code == DW_DLV_NO_ENTRY) {
	fprintf(stderr, "%s NO ENTRY:  %s: \n", program_name, msg);
    } else if (dwarf_code == DW_DLV_OK) {
	fprintf(stderr, "%s:  %s \n", program_name, msg);
    } else {
	fprintf(stderr, "%s InternalError:  %s:  code %d\n",
		program_name, msg, dwarf_code);
    }
    check_error = 1;
    longjmp(jmpbuf, 0);
}
