/* 
  Copyright (C) 2000,2004,2005,2006 Silicon Graphics, Inc.  All Rights Reserved.

  Copyright (C) 2006  NEC Corporation
  This file was modified from dwarfdump to analyze kernel debug information
  in makedumpfile.
  See below for the original copyright notice.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement
  or the like.  Any license provided herein, whether implied or
  otherwise, applies only to this software file.  Patent licenses, if
  any, provided herein do not apply to combinations of this program with
  other software, or any other product whatsoever.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston MA 02111-1307, USA.

  Contact information:  Silicon Graphics, Inc., 1500 Crittenden Lane,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan


$Header: /plroot/cmplrs.src/v7.4.5m/.RCS/PL/dwarfdump/RCS/print_die.c,v 1.51 2006/04/01 16:20:21 davea Exp $ */
#include "globals.h"
#include "dwarf_names.h"
#include "esb.h"		/* For flexible string buffer. */
#include "makename.h"		/* Non-duplicating string table. */
#include "dwarf_incl.h"

static void get_attr_value(Dwarf_Debug dbg, Dwarf_Half tag,
			   Dwarf_Attribute attrib, char **srcfiles,
			   Dwarf_Signed cnt
#ifdef DEBUG
			   , struct esb_s *esbp
#endif
			  );
static void print_attribute(Dwarf_Debug dbg, Dwarf_Die die,
			    Dwarf_Half attr,
			    Dwarf_Attribute actual_addr,
			    char **srcfiles,
			    Dwarf_Signed cnt);
static void get_location_list(Dwarf_Debug dbg, Dwarf_Die die,
			      Dwarf_Attribute attr
#ifdef DEBUG
			      , struct esb_s *esbp
#endif
			     );

#ifdef DEBUG
static struct esb_s esb_base;	/* static so gets initialized to zeros. 
				 */
#endif

static int indent_level = 0;
static boolean local_symbols_already_began = FALSE;

typedef string(*encoding_type_func) (Dwarf_Debug dbg, Dwarf_Half val);

/* Dwarf_Half list_of_attrs[] */
/*#include "at_list.i" unreferenced */

#define DIE_STACK_SIZE 50
static Dwarf_Die die_stack[DIE_STACK_SIZE];

#define PUSH_DIE_STACK(x) { die_stack[indent_level] = x; }
#define POP_DIE_STACK { die_stack[indent_level] = 0; }


/* recursively follow the die tree */
extern void
print_die_and_children(Dwarf_Debug dbg, Dwarf_Die in_die_in,
		       char **srcfiles, Dwarf_Signed cnt)
{
    Dwarf_Die child;
    Dwarf_Die sibling;
    Dwarf_Error err;
    int cdres = DW_DLV_OK;
    Dwarf_Die in_die = in_die_in;

    for (;;) {
	PUSH_DIE_STACK(in_die);

	/* here to pre-descent processing of the die */
	print_one_die(dbg, in_die, srcfiles, cnt);
	if (dwarf_info.status & DWARF_INFO_DONE)
	    goto exit_loop;

	cdres = dwarf_child(in_die, &child, &err);
	/* child first: we are doing depth-first walk */
	if (cdres == DW_DLV_OK) {
	    indent_level++;
	    print_die_and_children(dbg, child, srcfiles, cnt);
	    indent_level--;
	    if (indent_level == 0)
		local_symbols_already_began = FALSE;
	    dwarf_dealloc(dbg, child, DW_DLA_DIE);
	} else if (cdres == DW_DLV_ERROR) {
	    print_error(dbg, "dwarf_child", cdres, err);
	}

	cdres = dwarf_siblingof(dbg, in_die, &sibling, &err);
	if (cdres == DW_DLV_OK) {
	    /* print_die_and_children(dbg, sibling, srcfiles, cnt); We
	       loop around to actually print this, rather than
	       recursing. Recursing is horribly wasteful of stack
	       space. */
	} else if (cdres == DW_DLV_ERROR) {
	    print_error(dbg, "dwarf_siblingof", cdres, err);
	}

exit_loop:
	/* Here do any post-descent (ie post-dwarf_child) processing of 
	   the in_die. */

	POP_DIE_STACK;
	if (in_die != in_die_in) {
	    /* Dealloc our in_die, but not the argument die, it belongs 
	       to our caller. Whether the siblingof call worked or not. 
	     */
	    dwarf_dealloc(dbg, in_die, DW_DLA_DIE);
	}
	if (dwarf_info.status & DWARF_INFO_DONE)
	    break;
	if (cdres == DW_DLV_OK) {
	    /* Set to process the sibling, loop again. */
	    in_die = sibling;
	} else {
	    /* We are done, no more siblings at this level. */

	    break;
	}
    }				/* end for loop on siblings */
    return;
}

/* print info about die */
void
print_one_die(Dwarf_Debug dbg, Dwarf_Die die,
	      char **srcfiles, Dwarf_Signed cnt)
{
    Dwarf_Signed i;
    Dwarf_Off offset, overall_offset;
#ifdef DEBUG
    string tagname;
#endif
    Dwarf_Half tag;
    Dwarf_Signed atcnt;
    Dwarf_Attribute *atlist;
    int tres;
    int ores;
    int atres;
    unsigned long status_wk;

    if (dwarf_info.status & DWARF_INFO_DONE)
	return;
    tres = dwarf_tag(die, &tag, &err);
    dwarf_info.tmp_status = 0;
    if (tag == DW_TAG_structure_type) {
        dwarf_info.status = DWARF_INFO_TAG_STRUCT;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_TAG_STRUCT set 1\n"); /* debug */
#endif
    } else if (dwarf_info.status & DWARF_INFO_TAG_STRUCT && tag == DW_TAG_member) {
        dwarf_info.status |= DWARF_INFO_TAG_MEMBER;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_TAG_MEMBER set 1\n"); /* debug */
#endif
    } else {
        dwarf_info.status = 0;
        return;
    }
    if (tres != DW_DLV_OK) {
	print_error(dbg, "accessing tag of die!", tres, err);
    }
#ifdef DEBUG
    tagname = get_TAG_name(dbg, tag);
#endif
    ores = dwarf_dieoffset(die, &overall_offset, &err);
    if (ores != DW_DLV_OK) {
	print_error(dbg, "dwarf_dieoffset", ores, err);
    }
    ores = dwarf_die_CU_offset(die, &offset, &err);
    if (ores != DW_DLV_OK) {
	print_error(dbg, "dwarf_die_CU_offset", ores, err);
    }

	if (indent_level == 0) {
#ifdef DEBUG
		fprintf
		    (stderr, "\nCOMPILE_UNIT<header overall offset = %llu>:\n",
		     overall_offset - offset);
#endif
	} else if (local_symbols_already_began == FALSE &&
		   indent_level == 1) {
#ifdef DEBUG
	    fprintf(stderr, "\nLOCAL_SYMBOLS:\n");
#endif
	    local_symbols_already_began = TRUE;
	}
#ifdef DEBUG
	    fprintf(stderr, "<%d><%5llu>\t%s\n", indent_level, offset, tagname);
#endif

    atres = dwarf_attrlist(die, &atlist, &atcnt, &err);
    if (atres == DW_DLV_ERROR) {
	print_error(dbg, "dwarf_attrlist", atres, err);
    } else if (atres == DW_DLV_NO_ENTRY) {
	/* indicates there are no attrs.  It is not an error. */
	atcnt = 0;
    }


    dwarf_info.cur_member = DWARF_INFO_MAX;
    for (i = 0; i < atcnt; i++) {
	Dwarf_Half attr;
	int ares;

	ares = dwarf_whatattr(atlist[i], &attr, &err);
	if (ares == DW_DLV_OK) {
	    print_attribute(dbg, die, attr,
			    atlist[i],
			    srcfiles, cnt);
	} else {
	    print_error(dbg, "dwarf_whatattr entry missing", ares, err);
	}
    }
    if (dwarf_info.tmp_status & DWARF_INFO_NAME_MATCH) {
	if (dwarf_info.tmp_status & DWARF_INFO_OFFSET_DONE) {
	    /* member found */
	    dwarf_info.member[dwarf_info.cur_member].offset
	    = dwarf_info.member[DWARF_INFO_MAX].offset;
	    dwarf_info.status |= DWARF_INFO_MEMBER_DONE
			     << dwarf_info.cur_member * DWARF_INFO_STATUS_SHIFT;
	}
    } else {
	if (!(dwarf_info.tmp_status & DWARF_INFO_NAME_FOUND)
	    &&
	    dwarf_info.tmp_status & DWARF_INFO_OFFSET_DONE
	    &&
	    dwarf_info.location_member != DWARF_INFO_MAX) {
	    /* member with location only */
	    dwarf_info.member[dwarf_info.location_member].offset
	    = dwarf_info.member[DWARF_INFO_MAX].offset;
	    dwarf_info.status |= DWARF_INFO_MEMBER_DONE
		<< dwarf_info.location_member * DWARF_INFO_STATUS_SHIFT;
	}
    }

    for (i = 0; i < atcnt; i++) {
	dwarf_dealloc(dbg, atlist[i], DW_DLA_ATTR);
    }
    if (atres == DW_DLV_OK) {
	dwarf_dealloc(dbg, atlist, DW_DLA_LIST);
    }

    status_wk = DWARF_INFO_TAG_STRUCT | DWARF_INFO_TAG_MEMBER
		| DWARF_INFO_FILE_NAME_MATCH | DWARF_INFO_STRUCT_SIZE_DONE;
#ifdef DEBUG
	fprintf(stderr, "TAG_STRUCT,TAG_MEMBER,FILE_NAME_MATCH,STRUCT_SIZE_DONE set to wk\n"); /* debug */
#endif
    for (i = 0; i < dwarf_info.no_of_member; i++) {
	status_wk |= DWARF_INFO_MEMBER_DONE << i*DWARF_INFO_STATUS_SHIFT;
    }
    if ((dwarf_info.status & status_wk) == status_wk) {
	dwarf_info.status |= DWARF_INFO_DONE;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_DONE set 1\n"); /* debug */
#endif
    }
    return;
}

/* Encodings have undefined signedness. Accept either
   signedness.  The values are small (they are defined
   in the DWARF specification), so the
   form the compiler uses (as long as it is
   a constant value) is a non-issue.

   If string_out is non-NULL, construct a string output, either
   an error message or the name of the encoding.
   The function pointer passed in is to code generated
   by a script at dwarfdump build time. The code for
   the val_as_string function is generated
   from dwarf.h.  See <build dir>/dwarf_names.c

   If string_out is non-NULL then attr_name and val_as_string
   must also be non-NULL.

*/
int
get_small_encoding_integer_and_name(Dwarf_Debug dbg,
				    Dwarf_Attribute attrib,
				    Dwarf_Unsigned * uval_out,
				    char *attr_name,
				    string * string_out,
				    encoding_type_func val_as_string,
				    Dwarf_Error * err)
{
    Dwarf_Unsigned uval = 0;
    char buf[100];		/* The strings are small. */
    int vres = dwarf_formudata(attrib, &uval, err);

    if (vres != DW_DLV_OK) {
	Dwarf_Signed sval = 0;

	vres = dwarf_formsdata(attrib, &sval, err);
	if (vres != DW_DLV_OK) {
	    if (string_out != 0) {
		snprintf(buf, sizeof(buf),
			 "%s has a bad form.", attr_name);
		*string_out = buf;
	    }
	    return vres;
	}
	*uval_out = (Dwarf_Unsigned) sval;
    } else {
	*uval_out = uval;
    }
    if (string_out)
	*string_out = val_as_string(dbg, (Dwarf_Half) uval);

    return DW_DLV_OK;

}


static void
print_attribute(Dwarf_Debug dbg, Dwarf_Die die, Dwarf_Half attr,
		Dwarf_Attribute attr_in,
		char **srcfiles, Dwarf_Signed cnt)
{
    Dwarf_Attribute attrib = 0;
#ifdef DEBUG
    string atname = 0;
    string valname = 0;
#endif
    Dwarf_Half tag = 0;

    switch (attr) {
    case DW_AT_name:
	dwarf_info.tmp_status |= DWARF_INFO_NAME_FOUND;
    case DW_AT_byte_size:
    case DW_AT_decl_file:
    case DW_AT_data_member_location:
        break;
    default:
        return;
    }
#ifdef DEBUG
    atname = get_AT_name(dbg, attr);
#endif

    /* the following gets the real attribute, even in the ** face of an 
       incorrect doubling, or worse, of attributes */
    attrib = attr_in;
    /* do not get attr via dwarf_attr: if there are (erroneously) **
       multiple of an attr in a DIE, dwarf_attr will ** not get the
       second, erroneous one and dwarfdump ** will print the first one
       multiple times. Oops. */

    switch (attr) {
    case DW_AT_data_member_location:
#ifdef DEBUG
	esb_empty_string(&esb_base);
	get_location_list(dbg, die, attrib, &esb_base);
	valname = esb_get_string(&esb_base);
#else
	get_location_list(dbg, die, attrib);
#endif
	break;
    case DW_AT_name:
    case DW_AT_byte_size:
    case DW_AT_decl_file:
#ifdef DEBUG
	esb_empty_string(&esb_base);
	get_attr_value(dbg, tag, attrib, srcfiles, cnt, &esb_base);
	valname = esb_get_string(&esb_base);
#else
	get_attr_value(dbg, tag, attrib, srcfiles, cnt);
#endif
	break;
    default:
	break;
    }
#ifdef DEBUG
    fprintf(stderr, "\t\t%-28s%s\n", atname, valname);
#endif
}


static int
_dwarf_print_one_locdesc(Dwarf_Debug dbg,
			 Dwarf_Locdesc * llbuf
#ifdef DEBUG
			 , struct esb_s *string_out
#endif
			)
{

    Dwarf_Locdesc *locd;
    Dwarf_Half no_of_ops = 0;
#ifdef DEBUG
    string op_name;
#endif
    int i;

    locd = llbuf;
    no_of_ops = llbuf->ld_cents;
    for (i = 0; i < no_of_ops; i++) {
	Dwarf_Small op;
	Dwarf_Unsigned opd1;
#ifdef DEBUG
	Dwarf_Unsigned opd2;

	/* local_space_needed is intended to be 'more than big enough'
	   for a short group of loclist entries.  */
	char small_buf[100];

	if (i > 0)
	    esb_append(string_out, " ");
#endif

	op = locd->ld_s[i].lr_atom;
	if (op > DW_OP_nop) {
	    print_error(dbg, "dwarf_op unexpected value", DW_DLV_OK,
			err);
	    return DW_DLV_ERROR;
	}
#ifdef DEBUG
	op_name = get_OP_name(dbg, op);
	esb_append(string_out, op_name);
#endif

	opd1 = locd->ld_s[i].lr_number;
	if (dwarf_info.status & DWARF_INFO_TAG_MEMBER) {
	    dwarf_info.member[DWARF_INFO_MAX].offset
		= (int)locd->ld_s[i].lr_number;
	    dwarf_info.tmp_status |= DWARF_INFO_OFFSET_DONE;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_OFFSET_DONE set 1\n"); /* debug */
#endif
	}
#ifdef DEBUG
	if (op >= DW_OP_breg0 && op <= DW_OP_breg31) {
	    snprintf(small_buf, sizeof(small_buf),
		     "%+lld", (Dwarf_Signed) opd1);
	    esb_append(string_out, small_buf);
	} else {
	    switch (op) {
	    case DW_OP_addr:
		snprintf(small_buf, sizeof(small_buf), " %#llx", opd1);
		esb_append(string_out, small_buf);
		break;
	    case DW_OP_const1s:
	    case DW_OP_const2s:
	    case DW_OP_const4s:
	    case DW_OP_const8s:
	    case DW_OP_consts:
	    case DW_OP_skip:
	    case DW_OP_bra:
	    case DW_OP_fbreg:
		snprintf(small_buf, sizeof(small_buf),
			 " %lld", (Dwarf_Signed) opd1);
		esb_append(string_out, small_buf);
		break;
	    case DW_OP_const1u:
	    case DW_OP_const2u:
	    case DW_OP_const4u:
	    case DW_OP_const8u:
	    case DW_OP_constu:
	    case DW_OP_pick:
	    case DW_OP_plus_uconst:
	    case DW_OP_regx:
	    case DW_OP_piece:
	    case DW_OP_deref_size:
	    case DW_OP_xderef_size:
		snprintf(small_buf, sizeof(small_buf), " %llu", opd1);
		esb_append(string_out, small_buf);
		break;
	    case DW_OP_bregx:
		snprintf(small_buf, sizeof(small_buf), "%llu", opd1);
		esb_append(string_out, small_buf);
		opd2 = locd->ld_s[i].lr_number2;
		snprintf(small_buf, sizeof(small_buf),
			 "%+lld", (Dwarf_Signed) opd2);
		esb_append(string_out, small_buf);

		break;
	    default:
		break;
	    }
	}
#endif
    }

    return DW_DLV_OK;
}

/* Fill buffer with location lists 
   Buffer esbp expands as needed.
*/
 /*ARGSUSED*/ static void
get_location_list(Dwarf_Debug dbg, Dwarf_Die die, Dwarf_Attribute attr
#ifdef DEBUG
		  , struct esb_s *esbp
#endif
		  )
{
    Dwarf_Locdesc *llbuf = 0;
    Dwarf_Locdesc **llbufarray = 0;
    Dwarf_Signed no_of_elements;
    Dwarf_Error err;
    int i;
    int lres = 0;
    int llent = 0;

    lres = dwarf_loclist_n(attr, &llbufarray, &no_of_elements, &err);
    if (lres == DW_DLV_ERROR)
	print_error(dbg, "dwarf_loclist", lres, err);
    if (lres == DW_DLV_NO_ENTRY)
	return;

    for (llent = 0; llent < no_of_elements; ++llent) {
#ifdef DEBUG
	char small_buf[100];
#endif

	llbuf = llbufarray[llent];

#ifdef DEBUG
	if (llbuf->ld_from_loclist) {
	    if (llent == 0) {
		snprintf(small_buf, sizeof(small_buf),
			 "<loclist with %ld entries follows>",
			 (long) no_of_elements);
		esb_append(esbp, small_buf);
	    }
	    esb_append(esbp, "\n\t\t\t");
	    snprintf(small_buf, sizeof(small_buf), "[%2d]", llent);
	    esb_append(esbp, small_buf);
	}
#endif
#ifdef DEBUG
	lres = _dwarf_print_one_locdesc(dbg, llbuf, esbp);
#else
	lres = _dwarf_print_one_locdesc(dbg, llbuf);
#endif
	if (lres == DW_DLV_ERROR) {
	    return;
	} else {
	    /* DW_DLV_OK so we add follow-on at end, else is
	       DW_DLV_NO_ENTRY (which is impossible, treat like
	       DW_DLV_OK). */
	}
    }
    for (i = 0; i < no_of_elements; ++i) {
	dwarf_dealloc(dbg, llbufarray[i]->ld_s, DW_DLA_LOC_BLOCK);
	dwarf_dealloc(dbg, llbufarray[i], DW_DLA_LOCDESC);
    }
    dwarf_dealloc(dbg, llbufarray, DW_DLA_LIST);
}

/* Fill buffer with attribute value.
   We pass in tag so we can try to do the right thing with
   broken compiler DW_TAG_enumerator 

   We append to esbp's buffer.

*/
static void
get_attr_value(Dwarf_Debug dbg, Dwarf_Half tag, Dwarf_Attribute attrib,
	       char **srcfiles, Dwarf_Signed cnt
#ifdef DEBUG
	       , struct esb_s *esbp
#endif
	      )
{
    Dwarf_Half theform;
    string temps;
    Dwarf_Block *tempb;
    Dwarf_Signed tempsd = 0;
    Dwarf_Unsigned tempud = 0;
    int i;
    Dwarf_Half attr;
    Dwarf_Off off;
    Dwarf_Bool tempbool;
    Dwarf_Addr addr = 0;
    int fres;
    int bres;
    int wres;
    Dwarf_Half direct_form = 0;
#ifdef DEBUG
    char small_buf[100];
#endif


    fres = dwarf_whatform(attrib, &theform, &err);
    /* depending on the form and the attribute, process the form */
    if (fres == DW_DLV_ERROR) {
	print_error(dbg, "dwarf_whatform cannot find attr form", fres,
		    err);
    } else if (fres == DW_DLV_NO_ENTRY) {
	return;
    }

    dwarf_whatform_direct(attrib, &direct_form, &err);
    /* ignore errors in dwarf_whatform_direct() */


    switch (theform) {
    case DW_FORM_addr:
	bres = dwarf_formaddr(attrib, &addr, &err);
	if (bres == DW_DLV_OK) {
#ifdef DEBUG
	    snprintf(small_buf, sizeof(small_buf), "%#llx",
		     (unsigned long long) addr);
	    esb_append(esbp, small_buf);
#endif
	} else {
	    print_error(dbg, "addr formwith no addr?!", bres, err);
	}
	break;
    case DW_FORM_ref_addr:
	/* DW_FORM_ref_addr is not accessed thru formref: ** it is an
	   address (global section offset) in ** the .debug_info
	   section. */
	bres = dwarf_global_formref(attrib, &off, &err);
	if (bres == DW_DLV_OK) {
#ifdef DEBUG
	    snprintf(small_buf, sizeof(small_buf),
		     "<global die offset %llu>",
		     (unsigned long long) off);
	    esb_append(esbp, small_buf);
#endif
	} else {
	    print_error(dbg,
			"DW_FORM_ref_addr form with no reference?!",
			bres, err);
	}
	break;
    case DW_FORM_ref1:
    case DW_FORM_ref2:
    case DW_FORM_ref4:
    case DW_FORM_ref8:
    case DW_FORM_ref_udata:
	bres = dwarf_formref(attrib, &off, &err);
	if (bres != DW_DLV_OK) {
	    print_error(dbg, "ref formwith no ref?!", bres, err);
	}
	/* do references inside <> to distinguish them ** from
	   constants. */
#ifdef DEBUG
	snprintf(small_buf, sizeof(small_buf), "<%llu>", off);
	esb_append(esbp, small_buf);
#endif
	    wres = dwarf_whatattr(attrib, &attr, &err);
	    if (wres == DW_DLV_ERROR) {
		/* nothing */
	    } else if (wres == DW_DLV_NO_ENTRY) {
		/* nothing */
	    }
	break;
    case DW_FORM_block:
    case DW_FORM_block1:
    case DW_FORM_block2:
    case DW_FORM_block4:
	fres = dwarf_formblock(attrib, &tempb, &err);
	if (fres == DW_DLV_OK) {
	    for (i = 0; i < tempb->bl_len; i++) {
#ifdef DEBUG
		snprintf(small_buf, sizeof(small_buf), "%02x",
			 *(i + (unsigned char *) tempb->bl_data));
		esb_append(esbp, small_buf);
#endif
	    }
	    dwarf_dealloc(dbg, tempb, DW_DLA_BLOCK);
	} else {
	    print_error(dbg, "DW_FORM_blockn cannot get block\n", fres,
			err);
	}
	break;
    case DW_FORM_data1:
    case DW_FORM_data2:
    case DW_FORM_data4:
    case DW_FORM_data8:
	fres = dwarf_whatattr(attrib, &attr, &err);
	if (fres == DW_DLV_ERROR) {
	    print_error(dbg, "FORM_datan cannot get attr", fres, err);
	} else if (fres == DW_DLV_NO_ENTRY) {
	    print_error(dbg, "FORM_datan cannot get attr", fres, err);
	} else {
	    switch (attr) {
	    case DW_AT_byte_size:
	    case DW_AT_decl_file:
		wres = get_small_encoding_integer_and_name(dbg,
							   attrib,
							   &tempud,
							   /* attrname */
		    (char *) NULL,
							   /* err_string 
							    */ 
							   (char **)
							   NULL,
							   (encoding_type_func) 0,
							   &err);

		if (wres == DW_DLV_OK) {
#ifdef DEBUG
		    snprintf(small_buf, sizeof(small_buf), "%llu",
			     tempud);
		    esb_append(esbp, small_buf);
#endif
		    if (attr == DW_AT_byte_size) {
			if (dwarf_info.status & DWARF_INFO_TAG_STRUCT &&
			    !(dwarf_info.status & DWARF_INFO_TAG_MEMBER)) {
			    dwarf_info.struct_size = (int)tempud;
			    dwarf_info.status |= DWARF_INFO_STRUCT_SIZE_DONE;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_STRUCT_SIZE_DONE set 1\n"); /* debug */
#endif
			}
		    }
		    if (attr == DW_AT_decl_file) {
			if (srcfiles && tempud > 0 && tempud <= cnt) {
			    /* added by user request */
			    /* srcfiles is indexed starting at 0, but
			       DW_AT_decl_file defines that 0 means no
			       file, so tempud 1 means the 0th entry in
			       srcfiles, thus tempud-1 is the correct
			       index into srcfiles.  */
			    char *nameaddr;
			    char *fname = srcfiles[tempud - 1];
			    int namelen = strlen(dwarf_info.file_name);
			    int namelen_srcfiles = strlen(fname);
			    if (namelen_srcfiles <= namelen) {
				nameaddr = fname;
			    } else {
				nameaddr = fname + (namelen_srcfiles - namelen);
			    }
			    if (dwarf_info.status & DWARF_INFO_TAG_MEMBER) {
				if (!strcmp(nameaddr, dwarf_info.file_name)) {
				    /* Name of source file matches. */
				    dwarf_info.status |= DWARF_INFO_FILE_NAME_MATCH;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_FILE_NAME_MATCH set 1\n"); /* debug */
#endif
				} else {
				    dwarf_info.status &= ~(DWARF_INFO_TAG_STRUCT | DWARF_INFO_TAG_MEMBER | DWARF_INFO_FILE_NAME_MATCH);
#ifdef DEBUG
	fprintf(stderr, "TAG_STRUCT, TAG_MEMBER, FILE_NAME_MATCH reset 3\n"); /* debug */
#endif
				}
			    }
#ifdef DEBUG
			    esb_append(esbp, " ");
			    esb_append(esbp, fname);
#endif
			}
		    }
		} else {
		    print_error(dbg, "Cannot get encoding attribute ..",
				wres, err);
		}
		break;
	    default:
		wres = dwarf_formsdata(attrib, &tempsd, &err);
		if (wres == DW_DLV_OK) {
#ifdef DEBUG
		    snprintf(small_buf, sizeof(small_buf), "%lld",
			     tempsd);
		    esb_append(esbp, small_buf);
#endif
		} else if (wres == DW_DLV_NO_ENTRY) {
		    /* nothing? */
		} else {
		    print_error(dbg, "Cannot get formsdata..", wres,
				err);
		}
		break;
	    }
	}
	break;
    case DW_FORM_sdata:
	wres = dwarf_formsdata(attrib, &tempsd, &err);
	if (wres == DW_DLV_OK) {
#ifdef DEBUG
	    snprintf(small_buf, sizeof(small_buf), "%lld", tempsd);
	    esb_append(esbp, small_buf);
#endif
	} else if (wres == DW_DLV_NO_ENTRY) {
	    /* nothing? */
	} else {
	    print_error(dbg, "Cannot get formsdata..", wres, err);
	}
	break;
    case DW_FORM_udata:
	wres = dwarf_formudata(attrib, &tempud, &err);
	if (wres == DW_DLV_OK) {
#ifdef DEBUG
	    snprintf(small_buf, sizeof(small_buf), "%llu", tempud);
	    esb_append(esbp, small_buf);
#endif
	} else if (wres == DW_DLV_NO_ENTRY) {
	    /* nothing? */
	} else {
	    print_error(dbg, "Cannot get formudata....", wres, err);
	}
	break;
    case DW_FORM_string:
    case DW_FORM_strp:
	wres = dwarf_formstring(attrib, &temps, &err);
	if (wres == DW_DLV_OK) {
	    fres = dwarf_whatattr(attrib, &attr, &err);
	    if (fres == DW_DLV_ERROR) {
		print_error(dbg, "FORM_datan cannot get attr", fres, err);
	    } else if (fres == DW_DLV_NO_ENTRY) {
		print_error(dbg, "FORM_datan cannot get attr", fres, err);
	    } else {
		if (attr == DW_AT_name) {
		    if (dwarf_info.status & DWARF_INFO_TAG_STRUCT) {
			if (!(dwarf_info.status & DWARF_INFO_TAG_MEMBER)) {
			    if (strcmp(temps, dwarf_info.struct_name)) {
				dwarf_info.status &= ~DWARF_INFO_TAG_STRUCT;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_TAG_STRUCT reset 4\n"); /* debug */
#endif
			    }
			} else {
			    int i;
			    for (i = 0; i < dwarf_info.no_of_member; i++) {
				if (dwarf_info.member[i].name == NULL) {
				    continue;
				}
#ifdef DEBUG
	fprintf(stderr, "&dwarf_info.member[ %d ].name = %p\n", i, &dwarf_info.member[i].name[0]); /* debug */
#endif
				if (strcmp(temps, dwarf_info.member[i].name) == 0) {
				    dwarf_info.cur_member = i;
				    dwarf_info.tmp_status |= DWARF_INFO_NAME_MATCH;
#ifdef DEBUG
	fprintf(stderr, "DWARF_INFO_NAME_MATCH set 1\n"); /* debug */
#endif
				    break;
				}
			    }
			}
		    }
		}
	    }
#ifdef DEBUG
	    esb_append(esbp, temps);
#endif
	} else if (wres == DW_DLV_NO_ENTRY) {
	    /* nothing? */
	} else {
	    print_error(dbg, "Cannot get formstr/p....", wres, err);
	}

	break;
    case DW_FORM_flag:
	wres = dwarf_formflag(attrib, &tempbool, &err);
	if (wres == DW_DLV_OK) {
#ifdef DEBUG
	    if (tempbool) {
		snprintf(small_buf, sizeof(small_buf), "yes(%d)",
			 tempbool);
		esb_append(esbp, small_buf);
	    } else {
		snprintf(small_buf, sizeof(small_buf), "no");
		esb_append(esbp, small_buf);
	    }
#endif
	} else if (wres == DW_DLV_NO_ENTRY) {
	    /* nothing? */
	} else {
	    print_error(dbg, "Cannot get formflag/p....", wres, err);
	}
	break;
    case DW_FORM_indirect:
#ifdef DEBUG
	/* We should not ever get here, since the true form was
	   determined and direct_form has the DW_FORM_indirect if it is
	   used here in this attr. */
	esb_append(esbp, get_FORM_name(dbg, theform));
#endif
	break;
    default:
	print_error(dbg, "dwarf_whatform unexpected value", DW_DLV_OK,
		    err);
    }
}
